C*
      PROGRAM ETRS89
CC
CC PURPOSE    :  TRANSFORM COORDINATES GIVEN IN ITRFyy AT EPOCH TC
CC               INTO ETRFxx
CC
CC REMARKS    :  ---
CC
CC AUTHOR     :  M. ROTHACHER
CC
CC CREATED    :  19-Nov-1996
CC
CC CHANGES    :  04-Aug-1999 PF: ADAPTED CALL TO SR GETCOO (TIMCRD ADDED)
CC               15-Oct-2001 HB: SWITCH TO NEW MENU
CC               25-Sep-2002 HU: REMOVE I_ASTLIB
CC               04-Mar-2003 MR: CORRECT CALL FOR WTCOOR
CC               19-Nov-2003 RD: READ INP-FILENAME IN READINPF
CC               02-Jan-2004 HU: MAXSTA INCREASED FROM 200 TO 400
CC               28-Jun-2004 RD: USE MAXCRD FROM M_MAXDIM FOR MAXSTA
CC               14-Jun-2005 SS: USE TIMCRD (EPOCH VARIABLE REMOVED)
CC               21-Jun-2005 MM: LFNUM.inc REMOVED (LFNUMs IN M_BERN)
CC               23-Jun-2005 MM: IMPLICIT NONE AND DECLARATIONS ADDED
CC               10-Aug-2005 SS: INVERSE TRANSFORMATION
CC               27-Feb-2007 AG: CALL DEFCON WITH PARAMETER
CC               23-Sep-2010 RD: ENABLE CPU COUNTER
CC               04-Feb-2011 RD: USE GETCO3/WTSTAT INSTEAD OF GETCOO/WTCOOR
CC               17-Feb-2011 RD: COMMON MCMSTA NOT NEEDED ANYMORE
CC               14-Nov-2011 SL: NEW TITLE STRING FOR PRITIT, M_BERN WITH ONLY
CC               14-Dec-2011 SL: USE ITRF2[IE]TRF
CC               01-Jan-2017 ST: Use pgName
CC               18-Feb-2019 ST: Extract m_inpKey from d_inpKey because of
CC                               circular references created by usage of alcErr.
CC               24-Dec-2020 RD: New call of defcon
CC               28-Nov-2023 SL: Use startBernese instead of pritit etc.
CC               28-Nov-2023 SL: Print some more information to lfnPrt
CC               28-Nov-2023 SL: Support of velocity transformation
CC               31-Jan-2024 SL/SS: Adopting the ETRS89 tool extension made at PNAC
CC               01-Feb-2024 SL/SS: Refined calculation of dt (considering DOY)
CC               12-Mar-2024 SL/SS: Several improvements
CC               18-Feb-2025 SL/SS: Print PG ETRS89 instead of PGM ETRS89
CC
CC COPYRIGHT  :  ASTRONOMICAL INSTITUTE
CC      1996     UNIVERSITY OF BERN
CC               SWITZERLAND
CC
C*
      USE m_bern,   ONLY: i4b, r8b, shortLineLength, fileNameLength,
     1                    staNameLength, staFla2Length, staFlagLength,
     1                    timStrgLength, lfnPrt, lfnErr
      USE m_cpu,    ONLY: cpu_start
      USE o_bernese,ONLY: startBernese, stopBernese
      USE d_const,  ONLY: CONST
      USE d_datum,  ONLY: t_datum
      USE s_wtstat
      USE s_exitrc
      USE s_etrsin
      USE s_gtflna
      USE s_getco3
      USE s_getdat
      USE s_alcerr
      USE s_itrf2itrf
      USE s_itrf2etrf
      USE s_timst2
      USE s_pridat
      USE s_xyzell
      USE s_radgms
      USE s_eccell
      USE f_djul

      IMPLICIT NONE


C DECLARATIONS
C ------------
      CHARACTER(LEN=*),          PARAMETER      :: pgName = 'ETRS89'
      CHARACTER(LEN=fileNameLength)             :: CRDINP, VELINP
      CHARACTER(LEN=fileNameLength)             :: CRDOUT, VELOUT
      CHARACTER(LEN=shortLineLength)            :: TITNEW
      CHARACTER(LEN=timStrgLength)              :: tString
      CHARACTER(LEN=staNameLength),
     1              POINTER,     DIMENSION(:)   :: CRDNAM, VELNAM
      CHARACTER(LEN=staNameLength),
     1              ALLOCATABLE, DIMENSION(:)   :: VELNA2
      CHARACTER(LEN=staFla2Length),
     1              POINTER,     DIMENSION(:)   :: CRDFLG, VELFLG
      CHARACTER(LEN=staFla2Length),
     1              ALLOCATABLE, DIMENSION(:)   :: VELFL2
      CHARACTER(LEN=8),
     1              POINTER,     DIMENSION(:)   :: SYSFLG
      CHARACTER(LEN=4),
     1              POINTER,     DIMENSION(:)   :: PLATE
      CHARACTER(LEN=4),
     1              ALLOCATABLE, DIMENSION(:)   :: PLAT2
      CHARACTER(LEN=staFlagLength),
     1                           DIMENSION(1)   :: FLAGS
      CHARACTER(LEN=1)                          :: SIGN1, SIGN2
      CHARACTER(LEN=100)                        :: FMT

      INTEGER(i4b), POINTER,     DIMENSION(:)   :: CRDNUM, VELNUM
      INTEGER(i4b), ALLOCATABLE, DIMENSION(:)   :: VELNU2
      INTEGER(i4b), ALLOCATABLE, DIMENSION(:)   :: C, V
      INTEGER(i4b)                              :: CRDIRC, VELIRC, IRC
      INTEGER(i4b)                              :: NFLAG
      INTEGER(i4b)                              :: NCRD, NVEL, NVE2
      INTEGER(i4b)                              :: ICRD, IVEL
      INTEGER(i4b)                              :: TR
      INTEGER(i4b)                              :: IAC
      INTEGER(i4b)                              :: I, J
      INTEGER(i4b)                              :: IDEG1, IDEG2
      INTEGER(i4b)                              :: IMIN1, IMIN2

      REAL(r8b),    POINTER,     DIMENSION(:,:) :: CRDXYZ, VELXYZ
      REAL(r8b),    ALLOCATABLE, DIMENSION(:,:) :: CRDNEW, VELNEW
      REAL(r8b),    ALLOCATABLE, DIMENSION(:,:) :: CRDHLP, VELHLP
      REAL(r8b),                 DIMENSION(3)   :: CRDELL, CRDEL2
      REAL(r8b),                 DIMENSION(3)   :: CRDECC, VELECC
      REAL(r8b),                 DIMENSION(8)   :: pari
      REAL(r8b),                 DIMENSION(7)   :: rati
      REAL(r8b),                 DIMENSION(7)   :: parc, parv
      REAL(r8b),                 DIMENSION(3,3) :: rotc, rotv
      REAL(r8b)                                 :: TIMCRD, doy, dt
      REAL(r8b)                                 :: SEC1, SEC2

      TYPE(t_datum)                             :: DATUM
      TYPE(t_datum)                             :: CRDDAT, VELDAT
      TYPE(t_datum)                             :: HLPDAT, ELLDAT


C START CPU COUNTER
C -----------------
      CALL cpu_start(.TRUE.)


C NULLIFY THE POINTERS
C --------------------
      NULLIFY(CRDNAM)
      NULLIFY(VELNAM)
      NULLIFY(CRDFLG)
      NULLIFY(VELFLG)
      NULLIFY(SYSFLG)
      NULLIFY(PLATE)

      NULLIFY(CRDNUM)
      NULLIFY(VELNUM)

      NULLIFY(CRDXYZ)
      NULLIFY(VELXYZ)


C WRITE TITLE AND FILE LIST
C -------------------------
      CALL startBernese(
     1     pgName, 'Transform coordinates to ETRS89 frame',
     1     rdConstFile = .TRUE.)


C READ INPUT OPTIONS (TRANSFORMATION PARAMETERS ...)
C --------------------------------------------------
      pari(:) = 0d0
      rati(:) = 0d0

      CALL ETRSIN(TITNEW,DATUM%name,pari,rati)


C GET INPUT FILE NAMES
C --------------------
      CALL GTFLNA(1,'COORD  ',CRDINP,CRDIRC)
      CALL GTFLNA(0,'VELOCI ',VELINP,VELIRC)


C READ ALL STATIONS FROM INPUT FILES
C ----------------------------------
      NFLAG    = 1
      FLAGS(1) = '@'

      NCRD = 0

      CALL GETCO3(CRDINP,NFLAG,FLAGS,NCRD,CRDNAM,STANUM=CRDNUM,
     1     STAFLG5=CRDFLG,XSTAT=CRDXYZ,DATUM=CRDDAT%name,
     1     TIMCRD=TIMCRD,SYSFLG=SYSFLG)

      IF(NCRD.EQ.0) THEN
        WRITE(lfnErr,"(/,' *** PG ',A,A)") pgName,
     1          ': No stations with coordinates found'
        WRITE(lfnErr,'(8X,A,A)') 'Coordinate file: ',CRDINP
        WRITE(lfnErr,*)
        CALL exitRc(2)
      ENDIF

      NVEL = 0

      IF(VELIRC.EQ.0) THEN

        CALL GETCO3(VELINP,NFLAG,FLAGS,NVEL,VELNAM,STANUM=VELNUM,
     1       STAFLG5=VELFLG,XSTAT=VELXYZ,DATUM=VELDAT%name,
     1       PLATE=PLATE)

        IF(NVEL.EQ.0) THEN
          WRITE(lfnErr,"(/,' *** PG ',A,A)") pgName,
     1            ': No stations with velocities found'
          WRITE(lfnErr,'(8X,A,A)') 'Velocitiy file: ',VELINP
          WRITE(lfnErr,*)
          CALL exitRc(2)
        ENDIF

      ENDIF


C CHECK DATUM CONSISTENCY (NOT MANUAL)
C -----------------------
      IF(NVEL>0.AND.
     1   ABS(pari(8))<=0d0.AND.
     1   VELDAT%name/=CRDDAT%name) THEN
        WRITE(lfnErr,"(/,' *** PG ',A,A)") pgName,
     1          ': Datum inconsistency found'
        WRITE(lfnErr,'(8X,A,A)') 'Datum coordinate file: ',CRDDAT%name
        WRITE(lfnErr,'(8X,A,A)') 'Datum velocity file:   ',VELDAT%name
        WRITE(lfnErr,*)
        CALL exitRc(2)
      ENDIF


C CHECK DATUM STRING FOR SUBROUTINE ITRF2ETRF (NOT MANUAL)
C -------------------------------------------
      IF(ABS(pari(8))<=0d0) THEN

        IF(CRDDAT%name(1:4) == 'ITRF'.OR.
     1     CRDDAT%name(1:3) == 'IGS' .OR.
     1     CRDDAT%name(1:3) == 'IGb') THEN
          HLPDAT%name = 'I'//DATUM%name(1:15)
          DATUM%name  = 'E'//DATUM%name(1:15)
          TR = 1
        ELSEIF(CRDDAT%name(1:4) == 'ETRF') THEN
          HLPDAT%name = 'I'//CRDDAT%name(2:16)
          DATUM%name  = 'I'//DATUM%name(1:15)
          TR = -1
        ELSE
          WRITE(lfnErr,"(/,' *** PG ',A,A)") pgName,
     1          ': Input datum is not supported'
          WRITE(lfnErr,'(8X,A,A)') 'Input datum: ',CRDDAT%name
          WRITE(lfnErr,*)
          CALL exitRc(2)
        ENDIF

      ENDIF


C GET THE CORRESPONDING VELOCITIES
C --------------------------------
      ALLOCATE(V(NCRD),STAT=IAC)
      CALL ALCERR(IAC,'V',(/NCRD/),pgName)

      DO ICRD=1,NCRD
        V(ICRD) = 0
      ENDDO

      NVE2 = 0

      IF(NVEL>0) THEN

        NVE2 = NVEL

        DO ICRD=1,NCRD
          DO IVEL=1,NVEL
            IF(CRDNAM(ICRD)==VELNAM(IVEL)) THEN
              V(ICRD) = IVEL
            ENDIF
          ENDDO
        ENDDO

        ALLOCATE(C(NVEL),STAT=IAC)
        CALL ALCERR(IAC,'C',(/NVEL/),pgName)

        DO IVEL=1,NVEL

          C(IVEL) = 0

          DO ICRD=1,NCRD
            IF(VELNAM(IVEL)==CRDNAM(ICRD)) THEN
              C(IVEL) = ICRD
            ENDIF
          ENDDO

          IF(C(IVEL).EQ.0) THEN
            WRITE(lfnErr,"(/,' ### PG ',A,A)") pgName,
     1            ': No coordinates found, station skipped'
            WRITE(lfnErr,'(8X,A,A)') 'Station: ',VELNAM(IVEL)
            WRITE(lfnErr,*)
            WRITE(lfnErr,*)

            NVE2 = NVE2 - 1
          ENDIF

        ENDDO

        IF(NVE2.EQ.0) THEN
          WRITE(lfnErr,"(/,' *** PG ',A,A)") pgName,
     1            ': No stations with corresponding velocities found'
          WRITE(lfnErr,*)
          CALL exitRc(2)
        ENDIF

C REMAP THE VELOCITY ARRAYS
        ALLOCATE(VELNA2(NVE2),STAT=IAC)
        CALL ALCERR(IAC,'VELNA2',(/NVE2/),pgName)
        ALLOCATE(VELFL2(NVE2),STAT=IAC)
        CALL ALCERR(IAC,'VELFL2',(/NVE2/),pgName)
        ALLOCATE(PLAT2(NVE2),STAT=IAC)
        CALL ALCERR(IAC,'PLAT2',(/NVE2/),pgName)
        ALLOCATE(VELNU2(NVE2),STAT=IAC)
        CALL ALCERR(IAC,'VELNU2',(/NVE2/),pgName)

        I = 0

        DO ICRD=1,NCRD
          DO IVEL=1,NVEL
            IF(CRDNAM(ICRD)==VELNAM(IVEL)) THEN
              I = I + 1
              VELNA2(I) = VELNAM(IVEL)
              VELFL2(I) = VELFLG(IVEL)
              PLAT2(I)  = PLATE(IVEL)
              VELNU2(I) = VELNUM(IVEL)
            ENDIF
          ENDDO
        ENDDO

      ENDIF


C ALLOCATE THE RESULT ARRAYS
C --------------------------
      ALLOCATE(CRDNEW(3,NCRD),STAT=IAC)
      CALL ALCERR(IAC,'CRDNEW',(/3,NCRD/),pgName)
      ALLOCATE(CRDHLP(3,NCRD),STAT=IAC)
      CALL ALCERR(IAC,'CRDHLP',(/3,NCRD/),pgName)

      IF(NVE2>0) THEN
        ALLOCATE(VELNEW(3,NVE2),STAT=IAC)
        CALL ALCERR(IAC,'VELNEW',(/3,NVE2/),pgName)
        ALLOCATE(VELHLP(3,NVE2),STAT=IAC)
        CALL ALCERR(IAC,'VELHLP',(/3,NVE2/),pgName)
      ENDIF


C EPOCH STRING
C ------------
      CALL timst2(1,1,TIMCRD,tString)

      WRITE(lfnPrt,*)
      WRITE(lfnPrt,'(1X,A)') 'TRANSFORMATION PARAMETERS'
      WRITE(lfnPrt,'(1X,25("-"),/)')

      WRITE(lfnPrt,'(1X,A,A)') 'Epoch of coordinates: ',tString
      WRITE(lfnPrt,*)


C PARAMETERS FROM PANEL
C ---------------------
      IF(ABS(pari(8))>0d0) THEN

        WRITE(lfnPrt,*)
        WRITE(lfnPrt,'(1X,A,A,A,A)')
     1        TRIM(CRDDAT%name),' to ',TRIM(DATUM%name),' (MANUAL):'
        I = LEN(TRIM(CRDDAT%name))+LEN(TRIM(DATUM%name))+13
        WRITE(FMT,'(A,I4,A)') '(1X,',I,'("-"))'
        WRITE(lfnPrt,FMT)
        WRITE(lfnPrt,*)

        WRITE(lfnPrt,'(1X,73("-"))')
        WRITE(lfnPrt,'(1X,3A8,A9,3A10,2X,A8)')
     1       'Tx','Ty','Tz',
     1       'D',
     1       'Rx','Ry','Rz',
     1       'Epoch'
        WRITE(lfnPrt,'(1X,3A8,A9,3A10,2X,A8)')
     1       'mm','mm','mm',
     1       'ppb',
     1       'mas','mas','mas',
     1       'year'
        WRITE(lfnPrt,'(1X,73("-"))')
        WRITE(lfnPrt,'(1X,3F8.2,F9.3,3F10.4,2X,F8.3)')
     1       pari(1),pari(2),pari(3),
     1       pari(4),
     1       pari(5),pari(6),pari(7),
     1       pari(8)
        WRITE(lfnPrt,'(1X,73("-"))')
        WRITE(lfnPrt,'(1X,3A8,A9,3A10)')
     1       'mm/y','mm/y','mm/y',
     1       'ppb/y',
     1       'mas/y','mas/y','mas/y'
        WRITE(lfnPrt,'(1X,73("-"))')
        WRITE(lfnPrt,'(1X,3F8.2,F9.3,3F10.4)')
     1       rati(1),rati(2),rati(3),
     1       rati(4),
     1       rati(5),rati(6),rati(7)
        WRITE(lfnPrt,'(1X,73("-"))')
        WRITE(lfnPrt,*)

        doy = (pari(8)-REAL(INT(pari(8))))*365.25d0+1d0
        dt  = (timcrd-DJUL(INT(pari(8)),1,doy))/365.25d0

        DO i=1,7
          IF(i==1.OR.i==2.OR.i==3) THEN
            parc(i) = (pari(i)+rati(i)*dt)*1d-3
            parv(i) = rati(i)*1d-3
          ELSEIF(i==4) THEN
            parc(i) = (pari(i)+rati(i)*dt)*1d-9
            parv(i) = rati(i)*1d-9
          ELSEIF(i==5.OR.i==6.OR.i==7) THEN
            parc(i) = (pari(i)+rati(i)*dt)*1d-3/CONST%ars
            parv(i) = rati(i)*1d-3/CONST%ars
          ENDIF
        ENDDO

        WRITE(lfnPrt,'(1X,A)')
     1       'Transformation parameters at epoch of coordinates:'
        WRITE(lfnPrt,'(1X,73("-"))')
        WRITE(lfnPrt,'(1X,3A8,A9,3A10,2X,A8)')
     1       'Tx','Ty','Tz',
     1       'D',
     1       'Rx','Ry','Rz',
     1       'DeltaT'
        WRITE(lfnPrt,'(1X,3A8,A9,3A10,2X,A8)')
     1       'mm','mm','mm',
     1       'ppb',
     1       'mas','mas','mas',
     1       'years'
        WRITE(lfnPrt,'(1X,73("-"))')
        WRITE(lfnPrt,'(1X,3F8.2,F9.3,3F10.4,2X,F8.3)')
     1       parc(1)*1d3,parc(2)*1d3,parc(3)*1d3,
     1       parc(4)*1d9,
     1       parc(5)*1d3*CONST%ars,parc(6)*1d3*CONST%ars,
     1       parc(7)*1d3*CONST%ars,dt
        WRITE(lfnPrt,'(1X,73("-"))')
        WRITE(lfnPrt,*)

        IF(ABS(parc(5)*1d3*CONST%ars)>250d0.OR.
     1     ABS(parc(6)*1d3*CONST%ars)>250d0.OR.
     1     ABS(parc(7)*1d3*CONST%ars)>250d0) THEN
          WRITE(lfnErr,"(/,' ### PG ',A,A)") pgName,
     1          ': Big rotation angles found!'
          WRITE(lfnErr,*)
          WRITE(lfnErr,*)
        ENDIF

      ENDIF


C TRANSFORMATION INTO NEW SYSTEM
C ------------------------------
      IVEL = 0

      DO ICRD=1,NCRD

        IF(V(ICRD)>0) IVEL = IVEL + 1

C USE PARAMETERS FROM PANEL
        IF(ABS(pari(8))>0d0) THEN

          rotc(1,1) =  parc(4)*CRDXYZ(1,ICRD)
          rotc(1,2) = -parc(7)*CRDXYZ(2,ICRD)
          rotc(1,3) =  parc(6)*CRDXYZ(3,ICRD)
          rotc(2,1) =  parc(7)*CRDXYZ(1,ICRD)
          rotc(2,2) =  parc(4)*CRDXYZ(2,ICRD)
          rotc(2,3) = -parc(5)*CRDXYZ(3,ICRD)
          rotc(3,1) = -parc(6)*CRDXYZ(1,ICRD)
          rotc(3,2) =  parc(5)*CRDXYZ(2,ICRD)
          rotc(3,3) =  parc(4)*CRDXYZ(3,ICRD)

          DO I=1,3
            CRDNEW(I,ICRD) = CRDXYZ(I,ICRD)
            DO J=1,3
              CRDNEW(I,ICRD) = CRDNEW(I,ICRD)+rotc(I,J)
            ENDDO
            CRDNEW(I,ICRD) = CRDNEW(I,ICRD)+parc(I)
          ENDDO

          IF(V(ICRD)>0) THEN

            rotv(1,1) =  parv(4)*CRDXYZ(1,ICRD)
            rotv(1,2) = -parv(7)*CRDXYZ(2,ICRD)
            rotv(1,3) =  parv(6)*CRDXYZ(3,ICRD)
            rotv(2,1) =  parv(7)*CRDXYZ(1,ICRD)
            rotv(2,2) =  parv(4)*CRDXYZ(2,ICRD)
            rotv(2,3) = -parv(5)*CRDXYZ(3,ICRD)
            rotv(3,1) = -parv(6)*CRDXYZ(1,ICRD)
            rotv(3,2) =  parv(5)*CRDXYZ(2,ICRD)
            rotv(3,3) =  parv(4)*CRDXYZ(3,ICRD)

            DO I=1,3
              VELNEW(I,IVEL) = VELXYZ(I,V(ICRD))
              DO J=1,3
                VELNEW(I,IVEL) = VELNEW(I,IVEL)+rotv(I,J)
              ENDDO
              VELNEW(I,IVEL) = VELNEW(I,IVEL)+parv(I)
            ENDDO

          ENDIF

        ELSE

C USE PARAMETERS FROM TECHNICAL NOTE
          IF(TR.EQ.1) THEN

            IF(V(ICRD)>0) THEN
              CALL ITRF2ITRF(TIMCRD,CRDDAT%name,DATUM%name,
     1             CRDXYZ(:,ICRD),CRDHLP(:,ICRD),
     1             XVEL=VELXYZ(:,V(ICRD)),XVENEW=VELHLP(:,IVEL))
              CALL ITRF2ETRF(TIMCRD,DATUM%name,TR,CRDHLP(:,ICRD),
     1             CRDNEW(:,ICRD),
     1             XVEL=VELHLP(:,IVEL),XVENEW=VELNEW(:,IVEL))
            ELSE
              CALL ITRF2ITRF(TIMCRD,CRDDAT%name,DATUM%name,
     1             CRDXYZ(:,ICRD),CRDHLP(:,ICRD))
              CALL ITRF2ETRF(TIMCRD,DATUM%name,TR,CRDHLP(:,ICRD),
     1             CRDNEW(:,ICRD))
            ENDIF

          ELSEIF(TR.EQ.-1) THEN

            IF(V(ICRD)>0) THEN
              CALL ITRF2ETRF(TIMCRD,CRDDAT%name,TR,CRDXYZ(:,ICRD),
     1             CRDHLP(:,ICRD),
     1             XVEL=VELXYZ(:,V(ICRD)),XVENEW=VELHLP(:,IVEL))
              CALL ITRF2ITRF(TIMCRD,CRDDAT%name,DATUM%name,
     1             CRDHLP(:,ICRD),CRDNEW(:,ICRD),
     1             XVEL=VELHLP(:,IVEL),XVENEW=VELNEW(:,IVEL))
            ELSE
              CALL ITRF2ETRF(TIMCRD,CRDDAT%name,TR,CRDXYZ(:,ICRD),
     1             CRDHLP(:,ICRD))
              CALL ITRF2ITRF(TIMCRD,CRDDAT%name,DATUM%name,
     1             CRDHLP(:,ICRD),CRDNEW(:,ICRD))
            ENDIF

          ENDIF

        ENDIF

      ENDDO

      WRITE(lfnPrt,*)
      WRITE(lfnPrt,*)
      WRITE(lfnPrt,"(1X,A,A)") 'COORDINATES AND VELOCITIES',
     1     ' IN THE INPUT AND OUTPUT FRAMES'
      WRITE(lfnPrt,"(1X,57('-'),/)")

      WRITE(lfnPrt,'(1X,107("-"))')
      WRITE(lfnPrt,'(1X,A16,1X,A5,4A17,1X,A16)')
     1     'Station name    ',
     1     'Flag ',
     1     'X (m)/(m/y)',
     1     'Y (m)/(m/y)',
     1     'Z (m)/(m/y)',
     1     '3-D (m)/(m/y)',
     1     'Frame           '
      WRITE(lfnPrt,'(1X,107("-"))')

      IVEL = 0

      DO ICRD=1,NCRD

        IF(V(ICRD)>0) THEN
          IVEL = IVEL + 1
        ELSEIF(NVE2.GT.0) THEN
          WRITE(lfnErr,"(/,' ### PG ',A,A)") pgName,
     1          ': No velocities found'
          WRITE(lfnErr,'(8X,A,A)') 'Station: ',CRDNAM(ICRD)
          WRITE(lfnErr,*)
          WRITE(lfnErr,*)
        ENDIF

        WRITE(lfnPrt,'(1X,A16,1X,A5,4F17.7,1X,A16,1X,A)')
     1        CRDNAM(ICRD),CRDFLG(ICRD),CRDXYZ(:,ICRD),
     1        DSQRT(CRDXYZ(1,ICRD)**2+CRDXYZ(2,ICRD)**2+
     1        CRDXYZ(3,ICRD)**2),
     1        CRDDAT%name,'#CRD-XYZ'

        IF(V(ICRD)>0) THEN
          WRITE(lfnPrt,'(1X,A16,1X,A5,4F17.7,1X,A16,1X,A)')
     1          VELNAM(V(ICRD)),VELFLG(V(ICRD)),VELXYZ(:,V(ICRD)),
     1          DSQRT(VELXYZ(1,V(ICRD))**2+VELXYZ(2,V(ICRD))**2+
     1          VELXYZ(3,V(ICRD))**2),
     1          VELDAT%name,'#VEL-XYZ'
        ENDIF

        IF(ABS(pari(8))<=0d0) THEN
          WRITE(lfnPrt,'(1X,A16,1X,A5,4F17.7,1X,A16,1X,A)')
     1          CRDNAM(ICRD),CRDFLG(ICRD),CRDHLP(:,ICRD),
     1          DSQRT(CRDHLP(1,ICRD)**2+CRDHLP(2,ICRD)**2+
     1          CRDHLP(3,ICRD)**2),
     1          HLPDAT%name,'#CRD-XYZ'

          IF(V(ICRD)>0) THEN
            WRITE(lfnPrt,'(1X,A16,1X,A5,4F17.7,1X,A16,1X,A)')
     1            VELNA2(IVEL),VELFL2(IVEL),VELHLP(:,IVEL),
     1            DSQRT(VELHLP(1,IVEL)**2+VELHLP(2,IVEL)**2+
     1            VELHLP(3,IVEL)**2),
     1            HLPDAT%name,'#VEL-XYZ'
          ENDIF
        ENDIF

        WRITE(lfnPrt,'(1X,A16,1X,A5,4F17.7,1X,A16,1X,A)')
     1       CRDNAM(ICRD),CRDFLG(ICRD),CRDNEW(:,ICRD),
     1       DSQRT(CRDNEW(1,ICRD)**2+CRDNEW(2,ICRD)**2+
     1       CRDNEW(3,ICRD)**2),
     1       DATUM%name,'#CRD-XYZ'

        IF(V(ICRD)>0) THEN
          WRITE(lfnPrt,'(1X,A16,1X,A5,4F17.7,1X,A16,1X,A)')
     1          VELNA2(IVEL),VELFL2(IVEL),VELNEW(:,IVEL),
     1          DSQRT(VELNEW(1,IVEL)**2+VELNEW(2,IVEL)**2+
     1          VELNEW(3,IVEL)**2),
     1          DATUM%name,'#VEL-XYZ'
        ENDIF

        WRITE(lfnPrt,*)

      ENDDO


C LOCAL COORDINATE AND VELOCITY DIFFERENCES
C -----------------------------------------
      WRITE(lfnPrt,*)
      WRITE(lfnPrt,*)
      WRITE(lfnPrt,"(1X,A,A)")
     1     'COORDINATE AND VELOCITY DIFFERENCES',
     1     ' AS WELL AS THE LOCAL VELOCITIES'
      WRITE(lfnPrt,"(1X,67('-'),/)")

      ELLDAT%name = DATUM%name
      CALL GETDAT(ELLDAT,IRCODE=IRC)

      IF(IRC.EQ.1) THEN

        WRITE(lfnErr,"(/,' ### PG ',A,A)") pgName,
     1       ': Unsupported datum, local transformation with WGS - 84'
        WRITE(lfnErr,'(8X,A,A)') 'Datum: ',TRIM(ELLDAT%name)
        WRITE(lfnErr,*)
        WRITE(lfnErr,*)

        ELLDAT%name = 'WGS - 84'
        CALL GETDAT(ELLDAT)

      ENDIF

      CALL PRIDAT(ELLDAT)

      WRITE(lfnPrt,'(1X,133("-"))')
      WRITE(lfnPrt,'(1X,A16,1X,A5,4A17,1X,2A15,2X,A10)')
     1     'Station name    ',
     1     'Flag ',
     1     'X/N (m)/(m/y)',
     1     'Y/E (m)/(m/y)',
     1     'Z/U (m)/(m/y)',
     1     '3-D (m)/(m/y)',
     1     'Latitude',
     1     'Longitude',
     1     'Height (m)'
      WRITE(lfnPrt,'(1X,133("-"))')

      IVEL = 0

      DO ICRD=1,NCRD

        IF(V(ICRD)>0) IVEL = IVEL + 1

        IF(LEN_TRIM(CRDFLG(ICRD)).EQ.0) THEN

          WRITE(lfnErr,"(/,' ### PG ',A,A)") pgName,
     1          ': No CRD flag found, local transformations skipped'
          WRITE(lfnErr,'(8X,A,A)') 'Station: ',CRDNAM(ICRD)
          WRITE(lfnErr,*)
          WRITE(lfnErr,*)

          CYCLE

        ENDIF

        CALL XYZELL(ELLDAT,CRDNEW(:,ICRD),CRDELL)
        CALL RADGMS(1,CRDELL(1),SIGN1,IDEG1,IMIN1,SEC1)
        CALL RADGMS(1,CRDELL(2),SIGN2,IDEG2,IMIN2,SEC2)

C Latitude on a sphere for ECCELL
        CRDEL2 = CRDELL
        CRDEL2(1) = DATAN2(CRDNEW(3,ICRD),
     1       DSQRT(CRDNEW(1,ICRD)**2+CRDNEW(2,ICRD)**2))

        CRDHLP(:,ICRD) = CRDNEW(:,ICRD)-CRDXYZ(:,ICRD)

        CALL ECCELL(CRDEL2,CRDHLP(:,ICRD),CRDECC)

        WRITE(lfnPrt,'(1X,A16,1X,A5,4F17.7,1X,2F15.9,2X,F10.4,1X,A)')
     1        CRDNAM(ICRD),
     1        CRDFLG(ICRD),
     1        CRDHLP(:,ICRD),
     1        DSQRT(CRDHLP(1,ICRD)**2+CRDHLP(2,ICRD)**2+
     1        CRDHLP(3,ICRD)**2),
     1        SIGN(IDEG1+(IMIN1+SEC1/60)/60,CRDELL(1)),
     1        SIGN(IDEG2+(IMIN2+SEC2/60)/60,CRDELL(2)),
     1        CRDELL(3),'#dCRD-XYZ'

        WRITE(lfnPrt,'(1X,A16,1X,A5,4F17.7,1X,2F15.9,2X,F10.4,1X,A)')
     1        CRDNAM(ICRD),
     1        CRDFLG(ICRD),
     1        CRDECC,
     1        DSQRT(CRDECC(1)**2+CRDECC(2)**2+CRDECC(3)**2),
     1        SIGN(IDEG1+(IMIN1+SEC1/60)/60,CRDELL(1)),
     1        SIGN(IDEG2+(IMIN2+SEC2/60)/60,CRDELL(2)),
     1        CRDELL(3),'#dCRD-NEU'

        IF(V(ICRD)>0) THEN

          CALL ECCELL(CRDEL2,VELXYZ(:,V(ICRD)),VELECC)

          WRITE(lfnPrt,'(1X,A16,1X,A5,4F17.7,1X,2F15.9,2X,F10.4,1X,A)')
     1          VELNA2(IVEL),
     1          VELFL2(IVEL),
     1          VELECC,
     1          DSQRT(VELECC(1)**2+VELECC(2)**2+VELECC(3)**2),
     1          SIGN(IDEG1+(IMIN1+SEC1/60)/60,CRDELL(1)),
     1          SIGN(IDEG2+(IMIN2+SEC2/60)/60,CRDELL(2)),
     1          CRDELL(3),'#VEL-NEU ('//TRIM(CRDDAT%name)//')'

          CALL ECCELL(CRDEL2,VELHLP(:,IVEL),VELECC)

          WRITE(lfnPrt,'(1X,A16,1X,A5,4F17.7,1X,2F15.9,2X,F10.4,1X,A)')
     1          VELNA2(IVEL),
     1          VELFL2(IVEL),
     1          VELECC,
     1          DSQRT(VELECC(1)**2+VELECC(2)**2+VELECC(3)**2),
     1          SIGN(IDEG1+(IMIN1+SEC1/60)/60,CRDELL(1)),
     1          SIGN(IDEG2+(IMIN2+SEC2/60)/60,CRDELL(2)),
     1          CRDELL(3),'#VEL-NEU ('//TRIM(HLPDAT%name)//')'

          CALL ECCELL(CRDEL2,VELNEW(:,IVEL),VELECC)

          WRITE(lfnPrt,'(1X,A16,1X,A5,4F17.7,1X,2F15.9,2X,F10.4,1X,A)')
     1          VELNA2(IVEL),
     1          VELFL2(IVEL),
     1          VELECC,
     1          DSQRT(VELECC(1)**2+VELECC(2)**2+VELECC(3)**2),
     1          SIGN(IDEG1+(IMIN1+SEC1/60)/60,CRDELL(1)),
     1          SIGN(IDEG2+(IMIN2+SEC2/60)/60,CRDELL(2)),
     1          CRDELL(3),'#VEL-NEU ('//TRIM(DATUM%name)//')'

          VELHLP(:,IVEL) = VELNEW(:,IVEL)-VELXYZ(:,V(ICRD))

          CALL ECCELL(CRDEL2,VELHLP(:,IVEL),VELECC)

          WRITE(lfnPrt,'(1X,A16,1X,A5,4F17.7,1X,2F15.9,2X,F10.4,1X,A)')
     1          VELNA2(IVEL),
     1          VELFL2(IVEL),
     1          VELHLP(:,IVEL),
     1          DSQRT(VELHLP(1,IVEL)**2+VELHLP(2,IVEL)**2+
     1          VELHLP(3,IVEL)**2),
     1          SIGN(IDEG1+(IMIN1+SEC1/60)/60,CRDELL(1)),
     1          SIGN(IDEG2+(IMIN2+SEC2/60)/60,CRDELL(2)),
     1          CRDELL(3),'#dVEL-XYZ'

          WRITE(lfnPrt,'(1X,A16,1X,A5,4F17.7,1X,2F15.9,2X,F10.4,1X,A)')
     1          VELNA2(IVEL),
     1          VELFL2(IVEL),
     1          VELECC,
     1          DSQRT(VELECC(1)**2+VELECC(2)**2+VELECC(3)**2),
     1          SIGN(IDEG1+(IMIN1+SEC1/60)/60,CRDELL(1)),
     1          SIGN(IDEG2+(IMIN2+SEC2/60)/60,CRDELL(2)),
     1          CRDELL(3),'#dVEL-NEU'

        ENDIF

        WRITE(lfnPrt,*)

      ENDDO


C WRITE OUTPUT FILES
C ------------------
      CALL GTFLNA(1,'COORDRS',CRDOUT,CRDIRC)
      CALL WTSTAT(0,CRDOUT,TITNEW,DATUM%name,NCRD,CRDNAM,
     1     CRDNEW,CRDNUM,CRDFLG,TIMCRD,SYSFLG=SYSFLG)

      IF(NVE2>0) THEN
        CALL GTFLNA(1,'VELOUT ',VELOUT,VELIRC)
        CALL WTSTAT(0,VELOUT,TITNEW,DATUM%name,NVE2,VELNA2,
     1        VELNEW,VELNU2,VELFL2,PLATE=PLAT2)
      ENDIF


C DEALLOCATE MEMORY AND NULLIFY POINTERS
C --------------------------------------
      DEALLOCATE(C,STAT=IAC)
      DEALLOCATE(V,STAT=IAC)

      DEALLOCATE(VELNA2,STAT=IAC)
      DEALLOCATE(VELFL2,STAT=IAC)
      DEALLOCATE(PLAT2,STAT=IAC)
      DEALLOCATE(VELNU2,STAT=IAC)
      DEALLOCATE(CRDNEW,STAT=IAC)
      DEALLOCATE(VELNEW,STAT=IAC)
      DEALLOCATE(CRDHLP,STAT=IAC)
      DEALLOCATE(VELHLP,STAT=IAC)

      IF(ASSOCIATED(CRDNAM)) NULLIFY(CRDNAM)
      IF(ASSOCIATED(VELNAM)) NULLIFY(VELNAM)
      IF(ASSOCIATED(CRDFLG)) NULLIFY(CRDFLG)
      IF(ASSOCIATED(VELFLG)) NULLIFY(VELFLG)
      IF(ASSOCIATED(SYSFLG)) NULLIFY(SYSFLG)
      IF(ASSOCIATED(PLATE))  NULLIFY(PLATE)

      IF(ASSOCIATED(CRDNUM)) NULLIFY(CRDNUM)
      IF(ASSOCIATED(VELNUM)) NULLIFY(VELNUM)

      IF(ASSOCIATED(CRDXYZ)) NULLIFY(CRDXYZ)
      IF(ASSOCIATED(VELXYZ)) NULLIFY(VELXYZ)


      CALL stopBernese()

      CALL exitRc(0)

      END PROGRAM ETRS89
